local _G = _G
local Cartographer3 = {}
_G.Cartographer3 = Cartographer3

Cartographer3.Utils = {}
local Cartographer3_Data
local Cartographer3_Data_mapHolder
local Cartographer3_Data_scrollFrame
local Cartographer3_Data_mapView
local Cartographer3_Data_ZONE_DATA
local Cartographer3_Data_CONTINENT_DATA
local L

do
	local rotateMinimap = GetCVar("rotateMinimap") == "1"
	hooksecurefunc("SetCVar", function(key, value)
		if key == "rotateMinimap" then
			rotateMinimap = value == "1"
		end
	end)
	
	local MINIMAP_ARROW
	do
		local function f(...)
			local best
			for i = 1, select('#', ...) do
				local v = select(i, ...)
				if not v:GetName() and v:GetObjectType() == "Model" then
					best = v
				end
			end
			return best
		end
		MINIMAP_ARROW = f(Minimap:GetChildren())
	end
	
	function Cartographer3.Utils.GetPlayerRotation()
		if rotateMinimap then
			return MiniMapCompassRing:GetFacing()
		else
			return -MINIMAP_ARROW:GetFacing()
		end
	end
end

function Cartographer3.Utils.IsMouseHovering(frame)
	if not frame or not frame:IsVisible() then
		return false
	end
	local x, y = GetCursorPosition()
	local scale = frame:GetEffectiveScale()
	x, y = x/scale, y/scale
	
	local left, bottom, width, height = frame:GetRect()
	
	return x >= left and y >= bottom and x <= left+width and y <= bottom+height
end

do
	local frame = CreateFrame("Frame", nil, UIParent)
	
	local timers = {}
	local timers_whenNotShownAlso = {}
	function Cartographer3.Utils.AddTimer(func, whenNotShownAlso)
		timers[func] = true
		if whenNotShownAlso then
			timers_whenNotShownAlso[func] = true
		end
	end

	function Cartographer3.Utils.RemoveTimer(func)
		timers[func] = nil
		timers_whenNotShownAlso[func] = nil
	end
	
	local tmp = {}
	function Cartographer3.Utils.OnUpdate(this, elapsed)
		local currentTime = GetTime()
		local mapHolder_shown = Cartographer3_Data_mapHolder:IsShown()
		for func in pairs(timers) do
			tmp[func] = true
		end
		for func in pairs(tmp) do
			if timers[func] and (mapHolder_shown or timers_whenNotShownAlso[timer]) then
				func(elapsed, currentTime)
			end
			tmp[func] = nil
		end
	end
	
	local events = {}
	function Cartographer3.Utils.AddEventListener(event, func)
		if not events[event] then
			frame:RegisterEvent(event)
			events[event] = {}
		end
		table.insert(events[event], func)
	end
	
	function Cartographer3.Utils.RemoveEventListener(event, func)
		local events_event = events[event]
		if events_event then
			for i = #events_event, 1, -1 do
				local v = events_event[i]
				if v == func then
					table.remove(events_event, i)
				end
			end
			if not events_event[1] then
				frame:UnregisterEvent(event)
				events[event] = nil
			end
		end
	end
	
	local tmp = {}
	function Cartographer3.Utils.OnEvent(this, event, ...)
		local events_event = events[event]
		for i, func in ipairs(events_event) do
			tmp[i] = func
		end
		for i = 1, #tmp do
			local func = tmp[i]
			tmp[i] = nil
			func(event, ...)
		end
	end
	frame:SetScript("OnEvent", Cartographer3.Utils.OnEvent)
	local function f()
		Cartographer3.Utils.RemoveEventListener("ADDON_LOADED", f)
		f = nil
		frame:SetScript("OnUpdate", function()
			frame:SetScript("OnUpdate", Cartographer3.Utils.OnUpdate)
		end)
	end
	Cartographer3.Utils.AddEventListener("ADDON_LOADED", f)
end

function Cartographer3.Utils.ZoomToCurrentZone()
	SetMapToCurrentZone()
	Cartographer3.Utils.ZoomToZone(Cartographer3_Data.currentMapTexture)
end

function Cartographer3.Utils.ZoomToZone(texture)
	local zoneData = Cartographer3_Data_ZONE_DATA[texture]
	if zoneData then
		local zoneLeft = zoneData.mapLeft
		local zoneWidth = zoneData.mapWidth
		local zoneBottom = zoneData.mapBottom
		local zoneHeight = zoneData.mapHeight
		Cartographer3.moveMap(zoneLeft + zoneWidth / 2, zoneBottom + zoneHeight / 2, math.min(Cartographer3_Data_scrollFrame:GetWidth() / zoneWidth, Cartographer3_Data_scrollFrame:GetHeight() / zoneHeight))
		return
	end
	local continentData = Cartographer3_Data_CONTINENT_DATA[texture]
	if continentData then
		local left = continentData.visibleLeft
		local right = continentData.visibleRight
		local bottom = continentData.visibleBottom
		local top = continentData.visibleTop
		Cartographer3.moveMap((left + right) / 2, (bottom + top) / 2, math.min(Cartographer3_Data_scrollFrame:GetWidth() / (right - left), Cartographer3_Data_scrollFrame:GetHeight() / (top - bottom)))
		return
	end
	local instanceTextureFrame = Cartographer3_Data.instanceTextureFrames[texture]
	if instanceTextureFrame then
		Cartographer3.moveMap(instanceTextureFrame.x*0.01, instanceTextureFrame.y*0.01, math.min(Cartographer3_Data_scrollFrame:GetWidth() / (instanceTextureFrame.w*0.01), Cartographer3_Data_scrollFrame:GetHeight() / (instanceTextureFrame.h*0.01)))
	end
end

function Cartographer3.Utils.GetScaledCursorPosition()
	local x, y = GetCursorPosition()
	local scale = UIParent:GetEffectiveScale()
	return x/scale, y/scale
end

local cos, sin = _G.cos, _G.sin
function Cartographer3.Utils.RotateCoordinate(angle, x, y)
	local A = cos(angle)
	local B = sin(angle)
	return x * A - y * B, x * B + y * A
end

function Cartographer3.Utils.RotateTexCoord(angle)
	local A = cos(angle)
	local B = sin(angle)
	local ULx, ULy = -0.5 * A - -0.5 * B, -0.5 * B + -0.5 * A
	local LLx, LLy = -0.5 * A - 0.5 * B, -0.5 * B + 0.5 * A
	local URx, URy = 0.5 * A - -0.5 * B, 0.5 * B + -0.5 * A
	local LRx, LRy = 0.5 * A - 0.5 * B, 0.5 * B + 0.5 * A
	return ULx+0.5, ULy+0.5, LLx+0.5, LLy+0.5, URx+0.5, URy+0.5, LRx+0.5, LRy+0.5
end

function Cartographer3.Utils.UnpackFloat(D, C, B, A)
	local negative = A >= 128 and -1 or 1
	local exponent = 2*(A%128) + math.floor(B/128) - 127
	local mantissa = 1 + (65536*(B%128) + 256*C + D) / 2^23
	return negative * 2^exponent * mantissa
end

do
	local function f()
		if InCombatLockdown() then
			return
		end
		Cartographer3.Utils.RemoveTimer(f)
		collectgarbage('collect')
	end
	function Cartographer3.Utils.CollectGarbageSoon()
		Cartographer3.Utils.AddTimer(f)
	end
end

local instanceTop, instanceLeft, instanceBottom, instanceRight, instanceRows = 0, 0, 0, 0, 0

function Cartographer3.Utils.ConvertUniverseCoordinateToZoneCoordinate(x, y)
	for continentID, currentContinentData in ipairs(Cartographer3_Data_CONTINENT_DATA) do
		local continentLeft = currentContinentData.visibleLeft
		local continentRight = currentContinentData.visibleRight
		local continentBottom = currentContinentData.visibleBottom
		local continentTop = currentContinentData.visibleTop
		if x >= continentLeft and x <= continentRight and y >= continentBottom and y <= continentTop then
			local x_grid = math.floor((x - continentLeft) / (continentRight - continentLeft) * 10)
			local y_grid = math.floor((y - continentBottom) / (continentTop - continentBottom) * 10)
			local num = x_grid*10 + y_grid
			local searchData = Cartographer3_Data.zoneSearchData[continentID][num]
			local bestZone
			if searchData then
				for _, zoneTexture in ipairs(searchData) do
					local zoneData = Cartographer3_Data_ZONE_DATA[zoneTexture]
			
					if x >= zoneData.highlightLeft and x <= zoneData.highlightRight and y >= zoneData.highlightBottom and y <= zoneData.highlightTop then
						bestZone = zoneTexture
						break
					end
				end
			end
		
			if bestZone then
				local zoneData = currentContinentData.zones[bestZone]
				local left = zoneData.mapLeft
				local bottom = zoneData.mapBottom
				local width = zoneData.mapWidth
				local height = zoneData.mapHeight
				return bestZone, (x - left) / width, 1 - (y - bottom) / height
			else
				local center = currentContinentData.mapCenter
				local middle = currentContinentData.mapMiddle
				local width = currentContinentData.width
				local height = currentContinentData.height
				local left = center - width/2
				local bottom = middle - height/2
				return currentContinentData.texture, (x - left) / width, 1 - (y - bottom) / height
			end
		end
	end
	
	local instanceLeft = Cartographer3_Data.instanceLeft
	local instanceRight = Cartographer3_Data.instanceRight
	local instanceBottom = Cartographer3_Data.instanceBottom
	local instanceTop = Cartographer3_Data.instanceTop
	
	if x > instanceLeft and x < instanceRight and y > instanceBottom and y < instanceTop then
		local instanceRows = Cartographer3_Data.instanceRows
		local instance_column = math.floor((x - instanceLeft) / (instanceRight - instanceLeft) * instanceRows)
		local instance_row = math.floor((y - instanceTop) / (instanceBottom - instanceTop) * instanceRows)
		local instance_id = instance_column + instance_row * instanceRows + 1
		local instanceTextureFrame = Cartographer3_Data.instanceTextureFrames[instance_id]
		if instanceTextureFrame then
			local mapScale = Cartographer3_Data_mapView:GetEffectiveScale()
			local mapX, mapY, mapW, mapH = Cartographer3_Data_mapView:GetRect()
		
			local frame_center = instanceTextureFrame.x * 0.01
			local frame_middle = instanceTextureFrame.y * 0.01
			local frame_width = instanceTextureFrame.w * 0.01
			local frame_height = instanceTextureFrame.h * 0.01
		
			local x_coord = (x - frame_center) / frame_width + 0.5
			local y_coord = (y - frame_middle) / frame_height + 0.5
			if x_coord >= 0 and x_coord <= 1 and y_coord >= 0 and y_coord <= 1 then
				return instanceTextureFrame.name, x_coord, y_coord
			end
		end
	end
	return nil, nil, nil
end

function Cartographer3.Utils.ConvertZoneCoordinateToUniverseCoordinate(zone, x, y)
	if (x == 0 and y == 0) or not x then
		return nil, nil
	end
	local data = Cartographer3_Data_ZONE_DATA[zone]
	if data then
		y = 1 - y
		x = x * data.mapWidth + data.mapLeft
		y = y * data.mapHeight + data.mapBottom
		return x, y
	end
	data = Cartographer3_Data_CONTINENT_DATA[zone]
	if data then
		y = 1 - y
		local width, height = data.width, data.height
		x = x * width + data.mapCenter - width/2
		y = y * height + data.mapMiddle - height/2
		local zone_z, zone_x, zone_y = Cartographer3.Utils.ConvertUniverseCoordinateToZoneCoordinate(x, y)
		local zoneData = Cartographer3_Data_ZONE_DATA[zone_z]
		if zoneData then
			return x - zoneData.visibleLeft + zoneData.mapLeft, y - zoneData.visibleBottom + zoneData.mapBottom
		end
		local instanceTextureFrame = Cartographer3_Data.instanceTextureFrames[zone_z]
		if instanceTextureFrame then
			return (x - 0.5)*instanceTextureFrame.w*0.01 + instanceTextureFrame.x*0.01, (y - 0.5)*instanceTextureFrame.w*0.01 + instanceTextureFrame.y*0.01
		end
		return x, y
	end
	-- TODO: instances
	
	return nil, nil
end

function Cartographer3.Utils.GetCursorUniverseCoordinate()
	local cursorX, cursorY = GetCursorPosition()
	local mapScale = Cartographer3_Data_mapView:GetEffectiveScale()
	cursorX, cursorY = cursorX/mapScale, cursorY/mapScale
	local mapX, mapY, mapW, mapH = Cartographer3_Data_mapView:GetRect()
	
	cursorX, cursorY = cursorX - mapX - mapW/2, cursorY - mapY - mapH/2
	return cursorX, cursorY
end

do
	local cx, cy
	local function _getCursorZonePosition(cx, cy)
		local zoneData = Cartographer3_Data.currentZoneData
		if zoneData then
			local left = zoneData.mapLeft
			local bottom = zoneData.mapBottom
			local width = zoneData.mapWidth
			local height = zoneData.mapHeight
			if cx >= left and cy >= bottom and cx <= left+width and cy <= bottom+height then
				local zx, zy = (cx - left) / width, 1 - (cy - bottom) / height
				if not Cartographer3_Data.CITIES[Cartographer3_Data.currentMapTexture] then
					local name = UpdateMapHighlight(zx, zy)
					local fileName = Cartographer3_Data.LOCALIZED_ZONE_TO_TEXTURE[name]
					if fileName then
						zoneData = Cartographer3_Data_ZONE_DATA[fileName]
						if zoneData then
							left = zoneData.mapLeft
							bottom = zoneData.mapBottom
							width = zoneData.mapWidth
							height = zoneData.mapHeight
							if cx >= left and cy >= bottom and cx <= left+width and cy <= bottom+height then
								local zx, zy = (cx - left) / width, 1 - (cy - bottom) / height
								return fileName, zx, zy
							end
						end
					end
				end
				return Cartographer3_Data.currentMapTexture, zx, zy
			end
		end
		return Cartographer3.Utils.ConvertUniverseCoordinateToZoneCoordinate(cx, cy)
	end
	local ret_zz, ret_zx, ret_zy
	function Cartographer3.Utils.GetCursorZonePosition()
		local cx, cy = Cartographer3.Utils.GetCursorUniverseCoordinate()
		if last_cx == cx and last_cy == cy then
			return ret_zz, ret_zx, ret_zy
		end
		ret_zz, ret_zx, ret_zy = _getCursorZonePosition(cx, cy)
		return ret_zz, ret_zx, ret_zy
	end
end

function Cartographer3.Utils.GetUnitUniverseCoordinate(unit)
	local x, y = GetPlayerMapPosition(unit)
	if x == 0 and y == 0 then
		return nil, nil
	end
	return Cartographer3.Utils.ConvertZoneCoordinateToUniverseCoordinate(Cartographer3_Data.currentMapTexture, x, y)
end

function Cartographer3.Utils.AddUnitDataToFullTooltip(unit)
	local title
	if unit == "player" then
		title = L["Player:"]
	elseif unit:match("^party") then
		title = L["Party:"]
	elseif unit:match("^raid") then
		title = L["Raid:"]
	end
	local name, realm = UnitName(unit)
	if realm and realm ~= "" then
		name = name .. "-" .. realm
	end
	WorldMapTooltip:AddDoubleLine(title, name)
	
	local className, classEnglish = UnitClass(unit)
	local classColor = RAID_CLASS_COLORS[classEnglish]
	WorldMapTooltip:AddDoubleLine(L["Class:"], className, nil, nil, nil, classColor.r, classColor.g, classColor.b)
	
	local level = UnitLevel(unit)
	local levelColor = GetDifficultyColor(level)
	WorldMapTooltip:AddDoubleLine(L["Level:"], level, nil, nil, nil, levelColor.r, levelColor.g, levelColor.b)
	
	WorldMapTooltip:AddDoubleLine(L["Race:"], UnitRace(unit))
	
	local guildName, guildRank, guildRankNum = GetGuildInfo(unit)
	if guildName then
		local playerGuildName, playerGuildRank, playerGuildRankNum = GetGuildInfo('player')
		local r, g, b = 1, 1, 0
		if playerGuildName == guildName then
			r, g, b = 0, 1, 0
		end
		
		WorldMapTooltip:AddDoubleLine(L["Guild:"], ("<%s>"):format(guildName), nil, nil, nil, r, g, b)
		
		if playerGuildName == guildName then
			local rankDiff = guildRankNum - playerGuildRankNum
			if rankDiff >= 2 then
				r, g, b = 0, 1, 0
			elseif rankDiff == 1 then
				r, g, b = 0.5, 1, 0
			elseif rankDiff == 0 then
				r, g, b = 1, 1, 0
			elseif rankDiff == -1 then
				r, g, b = 1, 0.5, 0
			else
				r, g, b = 1, 0, 0
			end
		end
		WorldMapTooltip:AddDoubleLine(L["Rank:"], guildRank, nil, nil, nil, r, g, b)
		
		-- TODO: Guild note
	end
	
	-- TODO: Distance
end

do
	local tmp = {}
	function Cartographer3.Utils.AddUnitDataToTooltipLine(unit)
		local title
		if unit == "player" then
			title = L["Player:"]
		elseif unit:match("^party") then
			title = L["Party:"]
		elseif unit:match("^raid") then
			title = L["Raid:"]
		end
		tmp[#tmp+1] = title
	
		tmp[#tmp+1] = " "
	
		local name, realm = UnitName(unit)
		local _, class = UnitClass(unit)
		local classColor = RAID_CLASS_COLORS[class]
		tmp[#tmp+1] = "|cff"
		tmp[#tmp+1] = ("%02x"):format(classColor.r*255)
		tmp[#tmp+1] = ("%02x"):format(classColor.g*255)
		tmp[#tmp+1] = ("%02x"):format(classColor.b*255)
		tmp[#tmp+1] = name
		tmp[#tmp+1] = "|r"
		if realm and realm ~= "" then
			tmp[#tmp+1] = "-"
			tmp[#tmp+1] = realm
		end
	
		tmp[#tmp+1] = " "
	
		local level = UnitLevel(unit)
		local levelColor = GetDifficultyColor(level)
	
		tmp[#tmp+1] = "|cff"
		tmp[#tmp+1] = ("%02x"):format(levelColor.r*255)
		tmp[#tmp+1] = ("%02x"):format(levelColor.g*255)
		tmp[#tmp+1] = ("%02x"):format(levelColor.b*255)
		tmp[#tmp+1] = level
		tmp[#tmp+1] = "|r"
	
		local s = table.concat(tmp)
		for i = 1, #tmp do
			tmp[i] = nil
		end
		WorldMapTooltip:AddLine(s)
	end
end

local function ADDON_LOADED(event)
	Cartographer3.Utils.RemoveEventListener("ADDON_LOADED", ADDON_LOADED)
	ADDON_LOADED = nil
	
	L = Cartographer3.L
	Cartographer3_Data = Cartographer3.Data
	Cartographer3_Data_mapHolder = Cartographer3_Data.mapHolder
	Cartographer3_Data_scrollFrame = Cartographer3_Data.scrollFrame
	Cartographer3_Data_mapView = Cartographer3_Data.mapView
	Cartographer3_Data_ZONE_DATA = Cartographer3_Data.ZONE_DATA
	Cartographer3_Data_CONTINENT_DATA = Cartographer3_Data.CONTINENT_DATA
end
Cartographer3.Utils.AddEventListener("ADDON_LOADED", ADDON_LOADED)
